
.clear_button button.bk-bs-btn.bk-bs-btn-success {
	background-color: #d2d4d8;
	border-color: #d2d4d8;
	color: white;
}

.export_button button.bk-bs-btn.bk-bs-btn-success {
	background-color: #d2d4d8;
	border-color: #d2d4d8;
	color: white;
}

.expand-button{
    padding-top: 1px;
    padding-left: 150px;
}

.content {
    font-size: 200px;
}

.seed-input {
    color: black;
    padding-top: 16px;
}

.annotation-input {
    padding-left: 30px;
    padding-bottom: 1px;
}

.annotation-div {
    display: inline;
}

.annotation-button {
    padding-left: 150px;
}

phrase {
    font-family: Verdana;
    display: inline;
    border:1px black solid;
    border-radius: 5px;
    padding: 3px;
    font-size:12px;
    background:#d0efb8;
    line-height:40px;
}

.bad-word {
    color: #ed380b;
}

.checkbox_group {
    padding-top: 15px;
    padding-bottom: 0px;
}

.annotate_checkbox {
    padding-bottom: 0px;
}

.phrases_list {
    padding-top: 0px;
}

.group-div {
    white-space: pre-wrap;
}