# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
from os import path
from pathlib import Path

LIBRARY_PATH = Path(path.realpath(__file__)).parent
LIBRARY_ROOT = LIBRARY_PATH.parent
LIBRARY_OUT = Path(Path.home()) / 'nlp-architect' / 'cache'
LIBRARY_DATASETS = LIBRARY_ROOT / 'datasets'
